/*****************************************************************************\
 *
 * Author: Sauer Vaclav
 * Email:  xsauer00<at>stud.fit.vutbr.cz
 *
 * Bachelors's thesis: Space Game Project Using Delta3D Simulation Engine 2014
 * 
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, author will kindly welcome
 * if you give him credit and keep his names with their source code
 * and inform him about any succesful project using this code.
 *
\*****************************************************************************/

#include "Game.h"

// ----- VARIABLES ------------------------------------------------------------

// ----- CONSTRUCTORS AND DESTRUCTORS -----------------------------------------

/// Game constructor
Game::Game(const std::string& configFilename) : Application(configFilename)
{
	m_IsMainMenuLoaded = false;
	SetDefaultWindow();

	//Create gui, plyer, laser shooter, effects manager
	m_Gui              = new Gui(GetCamera(), GetKeyboard(), GetMouse(), &m_Config);
	m_Player           = new Player();
	m_Level            = new Level(&m_Save, m_Gui);
	m_EffectManager    = new dtCore::EffectManager();
	m_Destructible     = new Destructible();

	Sound::GetInstance();
	//Get the Listener from the Audio Manager and attach it to the Camera:
	dtAudio::Listener* audioListener = dtAudio::AudioManager::GetInstance().GetListener();
	GetCamera()->AddChild(audioListener);

	//Initialize other stuff
	GetCamera()->SetClearColor(0, 0, 0, 0);
	dtCore::System::GetInstance(0)->SetUseFixedTimeStep(true);
	GetScene()->SetUserCollisionCallback(NearCollisionCallback, this);
	m_Level->PlayerDead = true;

	//register effect for laser hit
    m_EffectManager->AddDetonationTypeMapping("LaserHit", "gamedata/STAGEMap/Particles/particleHit.osg");
    m_EffectManager->AddDetonationTypeMapping("Explosion", "gamedata/STAGEMap/Particles/Explosion.osg");

	//Load main menu
	LoadMainMenuScene();

	//Register senders for messages
	dtCore::Scene * scene = GetScene();
	AddSender(scene);
	AddSender(m_Gui);
	AddSender(m_Player);
	AddSender(m_Level);
	m_Player->AddSender(scene);
	m_Gui->AddSender(scene);
	Shooter::GetInstance()->AddSender(scene);
	Shooter::GetInstance()->AddSender(m_Player);
}

///Game destructor
Game::~Game(void)
{

	//Unregister senders for messages
	dtCore::Scene * scene = GetScene();
	RemoveSender(scene);
	RemoveSender(m_Gui);
	RemoveSender(m_Player);
	RemoveSender(m_Level);
	m_Player->RemoveSender(scene);
	m_Gui->RemoveSender(scene);
	Shooter::GetInstance()->RemoveSender(scene);
	Shooter::GetInstance()->RemoveSender(m_Player);
	
	//Destroy gui and sound manager
	m_Gui->~Gui();
	Sound::GetInstance()->~Sound();
	Shooter::GetInstance()->~Shooter();

	//Clear map
	dtCore::Project::GetInstance().ClearAllContexts();
	dtCore::Project::GetInstance().CloseAllMaps(true);
}

// ----- OVERRIDES ---------------------------------------------------------
///Configure game
void Game::Config()
{
	LoadConfig();
	LoadSave();
}

///Called when a key is pressed.
bool Game::KeyPressed(const dtCore::Keyboard* keyboard, int key )
{
	//Give event to palyer so he can move
	m_Player->KeyChanged(key, true);

	switch(key)
	{
	case osgGA::GUIEventAdapter::KEY_Escape:
		{
			//Pause/unpause game
			if(m_IsMainMenuLoaded == false)
			{
				GetMouse()->SetPosition(0, 0);
				bool previousState = m_Gui->IsVisible();
				GetWindow()->SetShowCursor(!previousState);
				m_Gui->IsVisible(!previousState);
				dtCore::System::GetInstance(0)->SetPause(!previousState);
			}
			return true;
		}
#ifdef _DEBUG
	case osgGA::GUIEventAdapter::KEY_T:
		{
			dtCore::Transform transform;	
			m_Player->Model->GetTransform(transform);

			transform.Set(0, 0, 0, 0, 0, 0);
			m_Player->Model->SetTransform(transform);

			m_Player->Model->GetBodyWrapper()->SetLinearVelocity(osg::Vec3(0, 0, 0));
			m_Player->Model->GetBodyWrapper()->SetAngularVelocity(osg::Vec3(0, 0, 0));

			m_Player->Hull        = 100;
			m_Player->Shield      = 100;
			m_Player->EnginePower = 0;

			return true;
		}
#endif
	case osgGA::GUIEventAdapter::KEY_F1:
		{
			//Show/hide help
			m_Config["ShowHelp"] = !m_Config["ShowHelp"].asBool();
			SaveConfig();
			m_Gui->HelpVisible(m_Config["ShowHelp"].asBool());
			return true;
		}
	}
	return false;
}

///Called when a key is released.
bool Game::KeyReleased(const dtCore::Keyboard* keyboard, int key )
{
	//Give event to palyer so he can stop moving
	m_Player->KeyChanged(key, false);
	return false;
}

///Called when a mouse is dragged (button down).  Overwrite for custom functionality.
bool Game::MouseDragged(const dtCore::Mouse* mouse, float x, float y)
{
	//Forward to player when we are not in menu
	if(!GetWindow()->GetShowCursor())
	{
		m_Player->MouseMoved(x, y);
		GetMouse()->SetPosition(0, 0);

		//Shoot if left button is pressed
		if(mouse->GetButtonState(dtCore::Mouse::MouseButton::LeftButton) == true)
		{
			m_Player->MousePressed(dtCore::Mouse::MouseButton::LeftButton);
		}
	}

	return true;
}

///Called when a mouse scroll wheel moved.  Overwrite for custom functionality.
bool Game::MouseScrolled(const dtCore::Mouse* mouse, int delta)
{
	//Accelerate/decelerate depending on scroll direction
	if(delta > 0)
	{
		m_Player->KeyChanged(osgGA::GUIEventAdapter::KEY_R, true);
		return true;
	}
	else if (delta < 0)
	{
		m_Player->KeyChanged(osgGA::GUIEventAdapter::KEY_F, true);
		return true;
	}
	return false;
}

///Called when a mouse is moved.  Overwrite for custom functionality.
bool Game::MouseMoved (const dtCore::Mouse *mouse, float x, float y)
{
	//Forward to player when we are not in menu
	if(GetWindow()->GetShowCursor() == false)
	{
		m_Player->MouseMoved(x, y);
		GetMouse()->SetPosition(0, 0);
	}
	return true;
}

//Called when a mouse button is pressed.  Overwrite for custom functionality.
bool Game::MouseButtonPressed(const dtCore::Mouse* mouse, dtCore::Mouse::MouseButton button)
{
	//Forward to player when we are not in menu
	if(!GetWindow()->GetShowCursor())
	{
		m_Player->MousePressed(button);
	}
	return false;
}

//Recieve messages.
void Game::OnMessage(MessageData* data)
{
	Application::OnMessage(data);
		
	//Process collision
	if (data->message == "collision")
	{
		dtCore::Scene::CollisionData* cd = static_cast< dtCore::Scene::CollisionData* >( data->userData );

		std::string name1 = dynamic_cast<Base*>(cd->mBodies[0])->GetName().c_str();
		std::string name2 = dynamic_cast<Base*>(cd->mBodies[1])->GetName().c_str();
		
		if(name1 == "LaserShot")
		{
			cd->mBodies[0]->SetActive(false);
		}
		if (name2 == "LaserShot")
		{
			cd->mBodies[1]->SetActive(false);
		}
		
		if((name1 == "Player" && name2 == "LaserShot") ||(name1 == "LaserShot" && name2 == "Player"))
		{
			m_Player->Damage(10, Player::DamageType::LASER_SHOT);
		}
		else if(name1 == "Player" || name2 == "Player")
		{
			float damage = m_Player->Model->GetBodyWrapper()->GetLinearVelocity().length() - 20;				
			if(damage <= 0) return;
			m_Player->Damage(damage, Player::DamageType::COLLISION);
		}

	}
	//Process player movement on physics update
	else if (data->message == "physics_step")
	{
		//Update movement of entities
		if(m_Gui->IsVisible() == false)
		{
			m_Player->ProcessMovement();
		}
	}
	//Player killed
	else if (m_IsMainMenuLoaded == false && data->message == "GameOver")
	{		
		m_Save.clear();;

		//Create explosion
		dtCore::Transform trans;	
		m_Player->Model->GetTransform(trans);
		m_EffectManager->AddDetonation(trans.GetTranslation(), "Explosion", 1.0f);
		
		//Stop ship
		m_Player->EnginePower = 0.0;
		static_cast<osgParticle::RandomRateCounter *>(m_Player->EngineParticle->GetSingleLayer("Layer 0")->GetModularEmitter().getCounter())->setRateRange(0, 0);
		
		//Show menu wit game over message
		m_IsMainMenuLoaded = true;
		m_Gui->IsVisible(true);
		GetWindow()->SetShowCursor(true);
		m_Gui->ShowStoryText("GAME OVER");
		m_Level->PlayerDead = true;
	}
	//Processing of gui events
	else if (data->message == "GUI")
	{
		Gui::Event * e = static_cast<Gui::Event *>( data->userData );
		switch(e->Value)
		{
		case Gui::ButtonEvent::NEW_GAME:
			{
				//Reset everything and start new game
				GetMouse()->SetPosition(0, 0);
				GetWindow()->SetShowCursor(false);
				NewSave();
				m_IsMainMenuLoaded = false;
				LoadMission(MISSION::BASE);
				m_Player->Shield = m_Player->Hull = 100;
				m_Gui->IsVisible(false);
				m_Level->PlayerDead = false;
				m_Level->Reset();
				m_Level->Advance();
				break;
			}
		case Gui::ButtonEvent::SETTINGS_OK:
			{
				//Apply and save new settings
				bool val = false;
				SaveConfig();
				val = m_Config["Fullscreen"].asBool();
				GetWindow()->SetFullScreenMode(m_Config["Fullscreen"].asBool());
				val = m_Config["ShowHelp"].asBool();
				m_Gui->HelpVisible(m_Config["ShowHelp"].asBool());
				val = m_Config["InvertX"].asBool();
				m_Player->InvertX = m_Config["InvertX"].asBool();
				val = m_Config["InvertY"].asBool();
				m_Player->InvertY = m_Config["InvertY"].asBool();
				break;
			}
		case Gui::ButtonEvent::STORY_ADVANCE:
			{
				if(m_Player->Hull != 0)
				{
					m_Level->Advance();
				}
				break;
			}
		case Gui::ButtonEvent::CONTINUE_GAME:
			{
				if(m_Player->Hull != 0)
				{
					//Load scene and set player and story progress depending on save
					GetMouse()->SetPosition(0, 0);
					GetWindow()->SetShowCursor(false);
					m_IsMainMenuLoaded = false;
					LoadMission(MISSION::BASE);
					m_Player->Shield = m_Save["Shield"].asDouble();
					m_Player->Hull   = m_Save["Hull"].asDouble();
					
					dtCore::Transform trans;
					trans.Set(	m_Save["p1"].asFloat(),
								m_Save["p2"].asFloat(),
								m_Save["p3"].asFloat(),
								m_Save["r1"].asFloat(),
								m_Save["r2"].asFloat(),
								m_Save["r3"].asFloat());

					m_Player->Model->SetTransform(trans);

					m_Gui->IsVisible(false);
					m_Level->PlayerDead = false;
					m_Level->Advance();
					break;
				}
				break;
			}
		case Gui::ButtonEvent::EXIT_GAME:
			{
				SaveSave();
				dtABC::Application::GetInstance(0)->Quit();
				break;
			}
		}
		delete e;
	}
}

///Override for frame
void Game::Frame(const double deltaSimTime)
{
	Application::Frame(deltaSimTime);
}

///Override for postframe
void Game::PostFrame(const double deltaSimTime)
{
	Application::PostFrame(deltaSimTime);
}

///Override for preframe
void Game::PreFrame(const double deltaSimTime)
{
	Application::PreFrame(deltaSimTime);

	//Update level
	m_Level->Update();

	//Update player
	if(m_Gui->IsVisible() == false)
	{
		m_Player->Update(deltaSimTime);
		m_Gui->UpdatePlayer(m_Player->Hull, m_Player->Shield, m_Player->EnginePower);
	}
	
	//Move ship and camera in main menu for nice effect
	if(m_IsMainMenuLoaded == true)
	{
		//Initialize variables
		dtCore::Camera *cam = GetCamera();
		double          time = dtCore::System::GetInstance().GetSimulationTime();
		float           posX;
		float           posY;

		GetMouse()->GetPosition(posX, posY);

		//Move camera based on time
		dtCore::Transform position;
		cam->GetTransform(position);
		position.SetTranslation(posX / 5 + sin(time * 0.5) / 2, 0, posY / 5 + cos(time * 0.4) / 2);
		osg::Vec3 rot;
		rot = position.GetRotation();
		position.SetRotation(sin(time * 0.2), cos(time * 0.3), 0);
		cam->SetTransform(position);

		//Move player based on time
		m_Player->Model->GetTransform(position);
		position.SetRotation(105, 0, -25 + sin(time * 1) * 3);
		m_Player->Model->SetTransform(position);
	}

	//Remove objects destroyed during collision callback.
	while(m_ObjectsToRemove.size() > 0)
	{
		dtCore::Transformable* trans = m_ObjectsToRemove.at(m_ObjectsToRemove.size() - 1);
		GetScene()->RemoveChild(trans);
		trans->SetActive(false);
		trans->ClearCollisionGeometry();
		m_ObjectsToRemove.pop_back();
	}
}

// ----- METHODS -----------------------------------------------------------

///Load game configuration and apply it to running game
void Game::LoadConfig()
{
	//Read file
	std::ifstream in("config/config.json");
	std::stringstream buffer;
	buffer << in.rdbuf();
	std::string contents = std::string(buffer.str());
	Json::Reader reader;
	bool parsingSuccessful = reader.parse(contents, m_Config);
	in.close();

	//In case that something went wrong, create new configuration
	if(parsingSuccessful == false)
	{
		NewConfig();
		std::cout << "New config created.\n" ;
		SaveConfig();
	}

	//Set configuration to game
	GetWindow()->SetFullScreenMode(m_Config["Fullscreen"].asBool());
	m_Gui->HelpVisible(m_Config["ShowHelp"].asBool());
	m_Player->InvertX = m_Config["InvertX"].asBool();
	m_Player->InvertY = m_Config["InvertY"].asBool();
}

///Loads saved game
void Game::LoadSave()
{
	std::ifstream in("save/save.json");
	std::stringstream buffer;
	buffer << in.rdbuf();
	std::string contents = std::string(buffer.str());
	Json::Reader reader;
	bool parsingSuccessful = reader.parse(contents, m_Save);
	if(parsingSuccessful == false || m_Save["Dialog1"].asBool() == false)
	{
		m_Gui->ContinueEnabled(false);
	}
}

///Processes objects shot by laser
void Game::ProcessShot(dtCore::Transformable* shot, dtCore::Transformable* target)
{
	//Disable shot, so it does not hit anthing else (will be removed later by shooter cache)
	if(shot != NULL)
	{
		dtCore::Transform trans;
		shot->GetTransform(trans);

		shot->SetActive(false);
		shot->ClearCollisionGeometry();

		//Add particel effect
		m_EffectManager->AddDetonation(trans.GetTranslation(), "LaserHit", 0.2f);
	}

	//Damage player if he was hit
	if(target != NULL && target->GetName() == "Player")
	{
		m_Player->Damage(10, Player::DamageType::LASER_SHOT);
	}
	//otherwise damage object
	else if(target != NULL )
	{
		if(m_Destructible->DamageObject(10, target, GetScene()))
		{		
			m_ObjectsToRemove.push_back(target);
		}
	}
}

///Processes collicion and damages entities if possible
void Game::ProcessCollision(dtCore::Scene::CollisionData cd)
{
	
	dtABC::ProximityTrigger* tr = dynamic_cast<dtABC::ProximityTrigger*>(cd.mBodies[1]);
	if(tr == NULL)
	{
		tr = dynamic_cast<dtABC::ProximityTrigger*>(cd.mBodies[0]);
	}
	if(tr != NULL)
	{
		//Trigger triggered
		return;
	}

	float damage = cd.mDepth * 60.0;

	//Damage player
	if(damage <= 0) return;
	if(cd.mBodies[0]!= NULL && cd.mBodies[0]->GetName() == "Player")
	{
		m_Player->Damage(damage, Player::DamageType::COLLISION);
	}
	else if(cd.mBodies[1]!= NULL && cd.mBodies[1]->GetName() == "Player")
	{
		m_Player->Damage(damage, Player::DamageType::COLLISION);
	}
	
	//Damage other objects, but only if damage is high enough
	damage = damage * 0.5;
	if(damage > 10)
	{
		if(m_Destructible->DamageObject(damage * 0.5, cd.mBodies[0], GetScene()))
		{		
			m_ObjectsToRemove.push_back(cd.mBodies[0]);
		}
		if(m_Destructible->DamageObject(damage * 0.5, cd.mBodies[1], GetScene()))
		{		
			m_ObjectsToRemove.push_back(cd.mBodies[1]);
		}
	}
}

//Loads mission
void Game::LoadMission(MISSION mission)
{
	std::string map = "";
	switch(mission)
	{
	case MISSION::BASE:
		map = "Base";
		break;

	//case MISSION::FIRST:
	//	map = "FirstMission";
	//	break;

	}

	//Clear current scene
	dtCore::Project::GetInstance().CloseAllMaps();
	GetScene()->RemoveAllDrawables();	

	//Clear variables
	m_IsMainMenuLoaded = false;

	//Load new map
	dtCore::Map &myMap = dtABC::BaseABC::LoadMap(map);

	//Get players model, shield, engine,...
	GetScene()->SetGravity(0.0f, 0.0f, 0.0f);
	m_Player->Model          = m_Level->GetPlayer(myMap);
	m_Player->ShieldModel    = m_Level->GetPlayerShield(myMap);
	m_Player->EngineParticle = m_Level->GetPlayerEngineParticle(myMap);
	m_Level->LoadLevelObjects(myMap);
	m_Player->EnginePower    = 0.0;

	//Load destructible objects
	m_Destructible->RegisterDestructibleObjects(myMap);

	//Set engine and shiled child of players model (so it moves with him)
	m_Player->EngineParticle->GetParent()->RemoveChild(m_Player->EngineParticle);
	m_Player->Model->AddChild(m_Player->EngineParticle);
	m_Player->ShieldModel->GetParent()->RemoveChild(m_Player->ShieldModel);
	m_Player->Model->AddChild(m_Player->ShieldModel);

	//Set camera  FOV		
	GetCamera()->SetPerspectiveParams(90, GetCamera()->GetAspectRatio(), 3, 5000);
	
	//Add particle manager
	AddDrawable(m_EffectManager.get());
}

///Creates new save for player
void Game::NewSave()
{
	//Set save
	m_Save.clear();
	m_Save["Hull"]          = 100;
	m_Save["Shield"]        = 100;
}

///Creates new default game configuration
void Game::NewConfig()
{
	//Set save
	m_Config.clear();
	m_Config["Fullscreen"] = false;
	m_Config["InvertX"]    = false;
	m_Config["InvertY"]    = false;
	m_Config["ShowHelp"]   = true;
}

///Saves game configuration
void Game::SaveConfig()
{
  std::ofstream myfile;
  myfile.open ("config/config.json");
  myfile << m_Config.toStyledString();
  std::cout << "Saved config" << m_Config.toStyledString();
  myfile.close();
}

///Saves players progress
void Game::SaveSave()
{
	//Save position and rotation
	dtCore::Transform trans;
	osg::Vec3 position;
	osg::Vec3 rotation;

	m_Player->Model->GetTransform(trans);
	trans.Get(position, rotation);

	m_Save["p1"]   = position.x();
	m_Save["p2"]   = position.y();
	m_Save["p3"]   = position.z();

	m_Save["r1"]   = rotation.x();
	m_Save["r2"]   = rotation.y();
	m_Save["r3"]   = rotation.z();

	//Save health
	m_Save["Shield"] = m_Player->Shield;
	m_Save["Hull"]   = m_Player->Hull;

	//Save file to hard drive
	std::ofstream myfile;
	myfile.open ("save/save.json");
	myfile << m_Save.toStyledString();
	std::cout << "Saved game" << m_Save.toStyledString();
	myfile.close();
}

///Loads scene in main menu
void Game::LoadMainMenuScene()
{
	try
	{
		dtCore::Project::GetInstance().SetContext("gamedata/STAGEMap", true);
	}
	catch (dtCore::ProjectInvalidContextException e)
	{
		e.Print();
		return;
	}

	//Get map and load it into scene
	dtCore::Map &myMap = dtABC::BaseABC::LoadMap("MainMenuScene");

	typedef std::vector<dtCore::RefPtr<dtCore::ActorProxy>> ProxyVector;
	ProxyVector proxies;
	myMap.FindProxies(proxies, "*");
	ProxyVector::iterator end_iter = proxies.end();

	// process the proxies
	for (ProxyVector::iterator iter = proxies.begin(); iter != end_iter; ++iter)
	{
		if ((*iter)->GetName() == "Ship")
		{
			m_Player->Model = dynamic_cast<dtCore::Object*>((*iter)->GetActor());
		}
	}
	GetCamera()->SetPerspectiveParams(20, GetCamera()->GetAspectRatio(), 3, 5000);
	m_IsMainMenuLoaded = true;
}

///Sets properties of window with game
void Game::SetDefaultWindow()
{
	GetWindow()->SetWindowTitle("SpaceSimGame");

	dtCore::DeltaWin::Resolution res = GetWindow()->GetCurrentResolution();

	int WindowRatio = 10;
	int WidthPart   = res.width / WindowRatio;
	int HeightPart  = res.height / WindowRatio;

	GetWindow()->SetPosition(WidthPart * 2 - 30, HeightPart * 2 - 30, WidthPart * (WindowRatio - 2), HeightPart * (WindowRatio - 2));	
}

// ----- OTHER STUFF -------------------------------------------------------

///Callback for collision handling
void NearCollisionCallback (void *data, dGeomID o1, dGeomID o2)
{	
	if (data == 0 || o1 == 0 || o2 == 0)
		return;

	Game * game = static_cast<Game*>(data);
	dtCore::ODESpaceWrap* spaceWrap = game->GetScene()->GetPhysicsController()->GetSpaceWrapper();

	dtCore::Transformable* c1 = static_cast<dtCore::Transformable*>(dGeomGetData(o1));
	dtCore::Transformable* c2 = static_cast<dtCore::Transformable*>(dGeomGetData(o2));

	dContactGeom contactGeoms[8];

	int numContacts = dCollide(o1, o2, 8, contactGeoms, sizeof(dContactGeom));

	if (numContacts > 0 && c1 != 0 && c2 != 0)
	{
		dtCore::Scene::CollisionData cd;

		cd.mBodies[0] = c1;
		cd.mBodies[1] = c2;

		if(c1 != NULL && c1->GetName() == "LaserShot" && c1->GetActive() == true)
		{
			game->ProcessShot(c1, c2);
			return;
		}
		if(c2 != NULL && c2->GetName() == "LaserShot" && c2->GetActive() == true)
		{
			game->ProcessShot(c2, c1);
			return;
		}

		cd.mLocation.set(contactGeoms[0].pos[0], contactGeoms[0].pos[1], contactGeoms[0].pos[2]);
		cd.mNormal.set(contactGeoms[0].normal[0], contactGeoms[0].normal[1], contactGeoms[0].normal[2]);
		cd.mDepth = contactGeoms[0].depth;
		game->ProcessCollision(cd);

		if (c1 != 0 || c2 != 0)
		{
			dContact contact;

			for (int i = 0; i < numContacts; ++i)
			{
				contact.surface.mode       = dContactBounce;
				contact.surface.mu         = (dReal)10.0; //Coulomb friction coefficient (0 - dInfinity)
				contact.surface.bounce     = (dReal)0.3;  //Restitution parameter (0 notbouncy - 1 full bounciness)
				contact.surface.bounce_vel = (dReal)0.01; //The minimum incoming velocity necessary for bounce

				contact.geom = contactGeoms[i];

				// Make sure to call these both, because in the case of
				// Trigger, meaningful stuff happens even if the return
				// is false.
				bool contactResult1 = c1->FilterContact(&contact, c2);
				bool contactResult2 = c2->FilterContact(&contact, c1);

				if (contactResult1 && contactResult2)
				{
					// All this also should be in a virtual function.
					dtCore::Physical* p1 = dynamic_cast<dtCore::Physical*>(c1);
					dtCore::Physical* p2 = dynamic_cast<dtCore::Physical*>(c2);

					if (p1 != 0 || p2 != 0)
					{
						dJointID joint = dJointCreateContact(game->GetScene()->GetWorldID(),
						spaceWrap->GetContactJoinGroupID(),
						&contact);

						dJointAttach(joint,
						p1 != 0 && p1->DynamicsEnabled() ? p1->GetBodyID() : 0,
						p2 != 0 && p2->DynamicsEnabled() ? p2->GetBodyID() : 0);
					}
				}
			}
		}
	}
}