/*****************************************************************************\
 *
 * Author: Sauer Vaclav
 * Email:  xsauer00<at>stud.fit.vutbr.cz
 *
 * Bachelors's thesis: Space Game Project Using Delta3D Simulation Engine 2014
 * 
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, author will kindly welcome
 * if you give him credit and keep his names with their source code
 * and inform him about any succesful project using this code.
 *
\*****************************************************************************/


#include"player.h"

//Ctor
Player::Player()
{
	Hull   = 100;
	Shield = 100;
	EnginePower = 0;
	Model = NULL;

	m_MoveLeft                = false;
	m_MoveRight               = false;
	m_MoveUp                  = false;
	m_MoveDown                = false;
	m_RotateLeft              = false;
	m_RotateRight             = false;
	m_MouseHorizontalMovement = 0;
	m_MouseVerticalMovement   = 0;
	m_LastLaserShot			  = 0;
	m_ShieldVisible           = 0;
	InvertX                   = false;
	InvertY                   = false;
}

//Store reference to ship model
void Player::SetModel(dtCore::RefPtr<dtCore::Object> playerModel)
{
	Model = playerModel;
}

///Update of player
void Player::Update(const double deltaSimTime)
{
	//Recharge shield
	if(Shield < 100)
		Shield += deltaSimTime * 2;

	//Show or hide shield model depending on whether it shoud be visible
	if(ShieldModel->GetActive() == false)
	{
		if(m_ShieldVisible > dtCore::System::GetInstance().GetSimulationTime())
		{
			ShieldModel->SetActive(true);
		}
	}
	else
	{
		if(m_ShieldVisible < dtCore::System::GetInstance().GetSimulationTime())
		{
			ShieldModel->SetActive(false);
		}
	}
	
	MoveItems();
}

///Processing of mouse movement
void Player::MouseMoved(float horizontal, float vertical)
{
	//Clamp values (helps when framerate is low)
	if(horizontal < -0.3) horizontal = -0.3;
	if(horizontal >  0.3) horizontal =  0.3;
	if(vertical   < -0.3) vertical   = -0.3;
	if(vertical   >  0.3) vertical   =  0.3;

	//Apply values based on defined settings
	m_MouseHorizontalMovement += InvertX ? horizontal : -horizontal;
	m_MouseVerticalMovement   += InvertY ? vertical   : -vertical;
}

///Processing of pressed mouse button
void Player::MousePressed(dtCore::Mouse::MouseButton button)
{
	switch(button)
	{
	case dtCore::Mouse::MouseButton::LeftButton:
		{
			//Periodically shoot laser
			double currTime = dtCore::System::GetInstance().GetSimulationTime();
			if(m_LastLaserShot < currTime)
			{
				m_LastLaserShot = currTime + 0.2;
				dtCore::Transform trans;
				Model->GetTransform(trans);
				SendMessage("ShootLaser", &trans);
				Model->AddChild(Sound::GetInstance()->PlayLaserShot());
			}
			break;
		}
	case dtCore::Mouse::MouseButton::MiddleButton:
		{
			//Toggle engine power
			EnginePower = EnginePower == 0.0 ? 1.0 : 0.0;
			break;
		}
	case dtCore::Mouse::MouseButton::RightButton:
		{
			break;
		}
	}
}

///Processing of pressed key
void Player::KeyChanged(const int key, const bool value)
{	
	switch(key)
		{
			case osgGA::GUIEventAdapter::KEY_W:
				{ m_MoveUp  = value; break; } 
			case osgGA::GUIEventAdapter::KEY_S:
				{ m_MoveDown = value; break; } 
			case osgGA::GUIEventAdapter::KEY_A:
				{ m_MoveLeft     = value; break; }  
			case osgGA::GUIEventAdapter::KEY_D:
				{ m_MoveRight    = value; break; } 
			case osgGA::GUIEventAdapter::KEY_Q:
				{ m_RotateLeft   = value; break; }  
			case osgGA::GUIEventAdapter::KEY_E:
				{ m_RotateRight  = value; break; } 
			case osgGA::GUIEventAdapter::KEY_R:
				{
					//Ship engine power increase
					if(value == true)
					{
						//First acceleration is gentle
						if(EnginePower > -0.15 && EnginePower < 0.15)
						{
							EnginePower += 0.1;
						}
						//Others are higher
						else
						{
							EnginePower += 0.2;
						}

						//Clamp value
						if(EnginePower > 1.0)
							EnginePower = 1.0;
					}
					break;
				} 
			case osgGA::GUIEventAdapter::KEY_F:
				{
					//Ship engine power decrease
					if(value == true)
					{
						//Deceleration when moving slowly is gentle
						if(EnginePower > -0.15 && EnginePower < 0.15)
						{
							EnginePower -= 0.1;
						}
						//Others are higher
						else
						{
							EnginePower -= 0.2;
						}

						//Clamp value
						if (EnginePower < -0.2)
							EnginePower = -0.2;
						if(EnginePower > -0.001 && EnginePower < 0.001)
							EnginePower = 0;
					}
					break;
				} 
		}
}

///Moves player based on engine power and pressed keys / moved mouse
void Player::ProcessMovement()
	{

		if(Model == NULL)
			return;

		//Default values used to calculate forces
		float engineSpeed   = 400 * EnginePower;
		float strafeSpeed   = 150;
		float rotationSpeed = 20;

		float forwardSpeed    = 0;
		float horizontalSpeed = 0;
		float verticalSpeed   = 0;
		float xRotation       = 0;
		float yRotation       = 0;
		float zRotation       = 0;
		

		//Proces forward/backward movement
		forwardSpeed += engineSpeed;

		//Process strafing
		if(m_MoveLeft)
		{
			horizontalSpeed -= strafeSpeed;
		}
		if(m_MoveRight)
		{
			horizontalSpeed += strafeSpeed;
		}
		if(m_MoveUp)
		{
			verticalSpeed += strafeSpeed;
		}
		if(m_MoveDown)
		{
			verticalSpeed -= strafeSpeed;
		}

		//Apply forces for movement
		Model->GetBodyWrapper()->ApplyRelForce(osg::Vec3(horizontalSpeed, forwardSpeed, verticalSpeed));
		
		//Process rotation of the ship by mouse
		xRotation = rotationSpeed * m_MouseHorizontalMovement * 20.0;
		m_MouseHorizontalMovement = 0;
	
		yRotation = rotationSpeed * m_MouseVerticalMovement * 20.0;
		m_MouseVerticalMovement = 0;

		//Process rolling of ship
		if(m_RotateLeft)
		{
			zRotation -= rotationSpeed;
		}
		if(m_RotateRight)
		{
			zRotation += rotationSpeed;
		}

		//Apply forces for rotation
		Model->GetBodyWrapper()->ApplyRelTorque(osg::Vec3(yRotation, zRotation, xRotation));
	}

///Moves camera, shield and engine particle with player.
void Player::MoveItems()
{
		osg::Vec3         position;
		dtCore::Transform transform;		
		osg::Quat         playerRotation;

		//Get players position and rotation
		Model->GetTransform(transform);
		transform.Get(position , playerRotation);

		//Position camera
		dtCore::Camera *cam = dtABC::Application::GetInstance(0)->GetCamera();

		cam->GetTransform(transform);
		transform.Set(position + (playerRotation * osg::Vec3(0, -55, 4.5)) , playerRotation);
		cam->SetTransform(transform);
		
		//Position shield
		if(ShieldModel->GetActive() == true)
		{
			transform.SetTranslation(position);
			ShieldModel->SetTransform(transform);
		}

		//Position engine particle
		transform.SetTranslation(position + (playerRotation * osg::Vec3(0, -14, 0)));
		EngineParticle->SetTransform(transform);

		UpdateEngineParticle();
}

///Updates engine particle lifetime and counter based on engine power.
void Player::UpdateEngineParticle()
{
	dtCore::ParticleLayer& particleLayer = *EngineParticle->GetSingleLayer("Layer 0");
	particleLayer.GetParticleSystem().getDefaultParticleTemplate().setLifeTime(std::max(0.0001, (EnginePower) / 3));

	osgParticle::RandomRateCounter *counter = static_cast<osgParticle::RandomRateCounter *>(particleLayer.GetModularEmitter().getCounter());

	if(EnginePower == 0.0)
		counter->setRateRange(0, 0);
	else
		counter->setRateRange(50, 50);
}

///Method used to damage player
void Player::Damage(double value, Player::DamageType damageType)
{
	if(value == 0)
		return;

	//Apply collision damage only once in a while.
	//so player is not immediatelly destroed due to
	//subsequent collisions after the frst one
	if(damageType == Player::DamageType::COLLISION && m_ShieldVisible > dtCore::System::GetInstance().GetSimulationTime())
	{
		return;
	}
	
	//Activate shield
	if(Shield > 10.)
	{
		m_ShieldVisible = dtCore::System::GetInstance().GetSimulationTime() + 0.2;
	}

	//Prevent collision damage for small collisions
	if(damageType == Player::DamageType::COLLISION && value < 10 )
	{
		return;
	}

	//Aplly damage
	Shield -= value;

	if(Shield < 0)
	{
		Hull += Shield;
		Shield = 0;
		if(Hull < 0)
		{		
			SendMessage("GameOver");
		}
	}

	//Show shield for longer time if it is not depleted by this damage
	if(Shield > 0)
	{
		m_ShieldVisible = dtCore::System::GetInstance().GetSimulationTime() + 0.5;
	}
}